use snafu::prelude::*;
use tokio::runtime::Runtime;

use crate::{Config, Error};
use crate::api::{Torrent, TorrentList, TorrentTracker};
use crate::api::qbittorrent::AsyncApiClient;
use crate::error::ApiSnafu as IOError;
use crate::utils::blocking_runtime;

#[derive(Debug)]
pub struct ApiClient {
    pub rt: Runtime,
    pub api: AsyncApiClient,
}

impl ApiClient {
    #[allow(dead_code)]
    pub fn from_config(config: &Config) -> Result<ApiClient, Error> {
        Self::new(&config.format_host(), &config.qbittorrent.login, &config.qbittorrent.password)
    }

    /// Login into a qBittorrent backend and return a proper ApiClient instance
    #[allow(dead_code)]
    pub fn new(host: &str, login: &str, password: &str) -> Result<ApiClient, Error> {
        let rt = blocking_runtime().context(IOError)?;
        let api = rt.block_on(AsyncApiClient::new(host, login, password))?;
        Ok(ApiClient {
            rt,
            api,
        })
    }

    #[allow(dead_code)]
    pub fn add(&self, magnet: &str, paused: bool) -> Result<(), Error> {
        self.rt.block_on(self.api.add(magnet, paused))
    }

    #[allow(dead_code)]
    pub fn get(&self, hash: &str) -> Result<Option<Torrent>, Error> {
        self.rt.block_on(self.api.get(hash))
    }

    #[allow(dead_code)]
    // TODO: Workaround
    pub fn get_with_cached_list(&self, hash: &str, list: &TorrentList) -> Result<Option<Torrent>, Error> {
        self.rt.block_on(self.api.get_with_cached_list(hash, list))
    }

    #[allow(dead_code)]
    pub fn list(&self) -> Result<TorrentList, Error> {
        self.rt.block_on(self.api.list())
    }

    #[allow(dead_code)]
    pub fn add_tracker(&self, hash: &str, url: &str) -> Result<(), Error> {
        self.rt.block_on(self.api.add_tracker(hash, url))
    }

    #[allow(dead_code)]
    pub fn get_trackers(&self, hash: &str) -> Result<Vec<TorrentTracker>, Error> {
        self.rt.block_on(self.api.get_trackers(hash))
    }

    #[allow(dead_code)]
    pub fn remove_tracker(&self, hash: &str, url: &str) -> Result<(), Error> {
        self.rt.block_on(self.api.remove_tracker(hash, url))
    }
}
